use std::borrow::Cow;

use cosmic_text::{Attrs, Buffer, Color, Family, FontSystem, Metrics, Shaping, Style, Weight};
use jotdown::{Container, Event, ListKind};

pub use jotdown;
use range_map::RangeMap;

pub struct JotdownBufferIter<'a, T: Iterator<Item = Event<'a>>> {
    djot: T,
    indent: Vec<Indent>,
}

struct JotdownIntoBuffer<'a, 'b, T: Iterator<Item = Event<'a>>> {
    djot: &'b mut T,
    attrs: Attrs<'static>,
    metrics: Metrics,
    indent: &'b mut Vec<Indent>,
    image_url: Option<Cow<'a, str>>,
    link_start: usize,
    urls: Vec<(range_map::Range<usize>, &'a str)>,
    location: usize,
    added: bool,
}

#[derive(Default, Clone, Copy)]
pub struct Indent {
    pub modifier: Option<ListKind>,
    pub indent: f32,
}

pub const INDENT_AMOUNT: f32 = 18.0;

impl<'a, 'b, T: Iterator<Item = Event<'a>>> Iterator for JotdownIntoBuffer<'a, 'b, T> {
    type Item = (&'a str, Attrs<'static>);

    fn next(&mut self) -> Option<Self::Item> {
        while let Some(event) = self.djot.next() {
            match event {
                Event::LeftSingleQuote => {
                    self.added = true;
                    self.location += "‘".len();
                    return Some(("‘", self.attrs.clone()));
                }
                Event::LeftDoubleQuote => {
                    self.added = true;
                    self.location += "“".len();
                    return Some(("“", self.attrs.clone()));
                }
                Event::RightSingleQuote => {
                    self.added = true;
                    self.location += "’".len();
                    return Some(("’", self.attrs.clone()));
                }
                Event::RightDoubleQuote => {
                    self.added = true;
                    self.location += "”".len();
                    return Some(("”", self.attrs.clone()));
                }
                Event::Ellipsis => {
                    self.added = true;
                    self.location += "…".len();
                    return Some(("…", self.attrs.clone()));
                }
                Event::EmDash => {
                    self.added = true;
                    self.location += "—".len();
                    return Some(("—", self.attrs.clone()));
                }
                Event::EnDash => {
                    self.added = true;
                    self.location += "–".len();
                    return Some(("–", self.attrs.clone()));
                }
                Event::Softbreak | Event::NonBreakingSpace => {
                    self.added = true;
                    self.location += " ".len();
                    return Some((" ", self.attrs.clone()));
                }
                Event::Str(Cow::Borrowed(s)) | Event::Symbol(Cow::Borrowed(s)) => {
                    self.added = true;
                    self.location += s.len();
                    return Some((s, self.attrs.clone()));
                }
                Event::Start(container, _) => match container {
                    Container::Heading { level, .. } => {
                        self.metrics = Metrics::new(4.0 - level as f32, 4.0 - level as f32);
                    }
                    Container::Emphasis => self.attrs = self.attrs.style(Style::Italic),
                    Container::Strong => self.attrs = self.attrs.weight(Weight::BOLD),
                    Container::Verbatim => self.attrs = self.attrs.family(Family::Monospace),
                    Container::List { kind, .. } => self.indent.push(Indent {
                        indent: self.indent.last().copied().unwrap_or_default().indent
                            + INDENT_AMOUNT,
                        modifier: Some(kind),
                    }),
                    Container::Image(url, _) => self.image_url = Some(url),
                    Container::Link(_, _) => {
                        self.link_start = self.location + 1;
                        self.attrs = self.attrs.color(Color::rgb(96, 198, 233));
                    }
                    _ => {}
                },
                Event::End(container) => match container {
                    Container::Emphasis => self.attrs = self.attrs.style(Style::Normal),
                    Container::Strong => self.attrs = self.attrs.weight(Weight::NORMAL),
                    Container::Verbatim => self.attrs = self.attrs.family(Family::SansSerif),
                    Container::List { .. } => {
                        self.indent.pop();
                    }
                    Container::Heading { .. } | Container::Paragraph | Container::Image(_, _) => {
                        if self.added {
                            return None;
                        }
                    }
                    Container::Link(Cow::Borrowed(url), _) => {
                        self.urls
                            .push((range_map::Range::new(self.link_start, self.location), url));
                        self.attrs = self.attrs.color(Color::rgb(255, 255, 255));
                    }
                    _ => {}
                },
                Event::Hardbreak | Event::ThematicBreak(_) => {}
                Event::Blankline | Event::Escape | Event::FootnoteReference(_) => {}
                Event::Str(Cow::Owned(_)) | Event::Symbol(Cow::Owned(_)) => panic!(),
            }
        }

        return None;
    }
}

pub struct JotdownItem<'a> {
    pub indent: Indent,
    pub buffer: Vec<(&'a str, Attrs<'static>)>,
    pub metrics: Metrics,
    pub image_url: Option<Cow<'a, str>>,
    pub url_map: Option<RangeMap<usize, &'a str>>,
}

impl<'a> JotdownItem<'a> {
    pub fn make_buffer(
        &self,
        font_system: &mut FontSystem,
        width: f32,
        metrics: Metrics,
    ) -> Buffer {
        let mut buffer = Buffer::new(
            font_system,
            Metrics::new(
                metrics.font_size * self.metrics.font_size,
                metrics.line_height * self.metrics.line_height,
            ),
        );
        buffer.set_rich_text(
            font_system,
            self.buffer.iter().cloned(),
            Attrs::new().family(Family::SansSerif),
            Shaping::Advanced,
        );

        buffer.set_wrap(font_system, cosmic_text::Wrap::WordOrGlyph);
        buffer.set_size(font_system, width - self.indent.indent, f32::MAX);
        buffer.shape_until_scroll(font_system, false);
        buffer
    }
}

impl<'a, T: Iterator<Item = Event<'a>>> Iterator for JotdownBufferIter<'a, T> {
    type Item = JotdownItem<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut jot = JotdownIntoBuffer {
            djot: &mut self.djot,
            attrs: Attrs::new().family(Family::SansSerif),
            indent: &mut self.indent,
            image_url: None,
            metrics: Metrics::new(1.0, 1.0),
            added: false,
            link_start: 0,
            location: 0,
            urls: Vec::new(),
        };

        let buffer: Vec<(&str, Attrs<'static>)> = (&mut jot).collect();
        let image_url = jot.image_url;
        let urls = jot.urls;
        let added = jot.added;
        let metrics = jot.metrics;
        let indent = self.indent.last().copied().unwrap_or_default();
        if !added {
            return None;
        } else {
            return Some(JotdownItem {
                indent,
                url_map: if urls.is_empty() {
                    None
                } else {
                    RangeMap::try_from_iter(urls.into_iter()).ok()
                },
                buffer,
                image_url,
                metrics,
            });
        }
    }
}

pub fn jotdown_into_buffers<'a, T: Iterator<Item = Event<'a>>>(
    djot: T,
) -> JotdownBufferIter<'a, T> {
    JotdownBufferIter {
        djot,
        indent: Vec::new(),
    }
}
